@@ -7,7 +7,7 @@
 
 use super::{selector::Database, FaceRef, FontSelector};
 use crate::conv::{to_u32, to_usize};
-use std::collections::HashMap;
+use std::collections::hash_map::{Entry, HashMap};
 use std::path::{Path, PathBuf};
 use std::sync::{RwLock, RwLockReadGuard};
 use thiserror::Error;
@@ -21,8 +21,8 @@ enum FontError {
     #[error("font load error")]
     TtfParser(#[from] ttf_parser::FaceParsingError),
     #[cfg(all(not(feature = "harfbuzz_rs"), feature = "rustybuzz"))]
-    #[error("unknown font read error")]
-    UnknownLoadError,
+    #[error("invalid units per EM")]
+    UnitsPerEm,
     #[cfg(feature = "ab_glyph")]
     #[error("font load error")]
     AbGlyphFontError(#[from] ab_glyph::InvalidFont),
@@ -90,15 +90,17 @@ impl<'a> FaceStore<'a> {
     ///
     /// The `path` is to be stored; its contents are already loaded in `data`.
     fn new(path: PathBuf, data: &'a [u8], index: u32) -> Result<Self, FontError> {
+        let face = Face::from_slice(data, index)?;
+        #[cfg(all(not(feature = "harfbuzz_rs"), feature = "rustybuzz"))]
+        let rustybuzz = rustybuzz::Face::from_face(face.clone()).ok_or(FontError::UnitsPerEm)?;
         Ok(FaceStore {
             path,
             index,
-            face: Face::from_slice(data, index)?,
+            face,
             #[cfg(feature = "harfbuzz_rs")]
             harfbuzz: harfbuzz_rs::Face::from_bytes(data, index).into(),
             #[cfg(all(not(feature = "harfbuzz_rs"), feature = "rustybuzz"))]
-            rustybuzz: rustybuzz::Face::from_slice(data, index)
-                .ok_or(FontError::UnknownLoadError)?,
+            rustybuzz,
             #[cfg(feature = "ab_glyph")]
             ab_glyph: ab_glyph::FontRef::try_from_slice_and_index(data, index)?,
             #[cfg(feature = "fontdue")]
@@ -131,14 +133,15 @@ impl FaceList {
 
 #[derive(Default)]
 struct FontList {
-    fonts: Vec<(FontId, Vec<FaceId>)>,
+    // A "font" is a list of faces (primary + fallbacks); we cache glyph-lookups per char
+    fonts: Vec<(FontId, Vec<FaceId>, HashMap<char, Option<FaceId>>)>,
     sel_hash: Vec<(u64, FontId)>,
 }
 
 impl FontList {
     fn push(&mut self, list: Vec<FaceId>, sel_hash: u64) -> FontId {
         let id = FontId(to_u32(self.fonts.len()));
-        self.fonts.push((id, list));
+        self.fonts.push((id, list, HashMap::new()));
         self.sel_hash.push((sel_hash, id));
         id
     }
@@ -162,6 +165,11 @@ pub struct FontLibrary {
 
 /// Font management
 impl FontLibrary {
+    /// Get a reference to the font database
+    pub fn read_db(&self) -> RwLockReadGuard<Database> {
+        self.db.read().unwrap()
+    }
+
     /// Get mutable access to the font database
     ///
     /// This can be used to adjust font selection. Note that any changes only
@@ -182,7 +190,7 @@ impl FontLibrary {
     /// (default) one.
     pub fn first_face_for(&self, font_id: FontId) -> FaceId {
         let fonts = self.fonts.read().unwrap();
-        for (id, list) in &fonts.fonts {
+        for (id, list, _) in &fonts.fonts {
             if *id == font_id {
                 return *list.first().unwrap();
             }
@@ -203,22 +211,35 @@ impl FontLibrary {
     ///
     /// This selects the first face containing this glyph from the font list.
     pub fn face_for_char(&self, font_id: FontId, c: char) -> Option<FaceId> {
-        let faces = self.faces.read().unwrap();
-        let fonts = self.fonts.read().unwrap();
-        let list = fonts
+        // TODO: `face.glyph_index` is a bit slow to use like this where several
+        // faces may return no result before we find a match. Caching results
+        // in a HashMap helps. Perhaps better would be to (somehow) determine
+        // the script/language in use and check whether the font face supports
+        // that, perhaps also checking it has shaping support.
+        let mut fonts = self.fonts.write().unwrap();
+        let font = fonts
             .fonts
-            .iter()
+            .iter_mut()
             .find(|item| item.0 == font_id)
-            .map(|item| &item.1)
             .expect("invalid FontId");
-        for face_id in list.iter() {
-            let face = &faces.faces[face_id.get()];
-            // TODO: should we only return faces with shaping data, somehow?
-            if face.face.glyph_index(c).is_some() {
-                return Some(*face_id);
+
+        match font.2.entry(c) {
+            Entry::Occupied(entry) => *entry.get(),
+            Entry::Vacant(entry) => {
+                let faces = self.faces.read().unwrap();
+                let list = &font.1;
+                let mut id: Option<FaceId> = None;
+                for face_id in list.iter() {
+                    let face = &faces.faces[face_id.get()];
+                    if face.face.glyph_index(c).is_some() {
+                        id = Some(*face_id);
+                        break;
+                    }
+                }
+                entry.insert(id);
+                id
             }
         }
-        None
     }
 
     /// Resolve the face for a character
@@ -233,14 +254,16 @@ impl FontLibrary {
 
     /// Select the default font
     ///
-    /// If `FontId(0)` has not been defined yet, this sets the default font,
-    /// otherwise it does nothing.
+    /// If the font database has not yet been initialised, it is initialised.
     ///
-    /// This *must* be called (at least once) before any other font-loading
+    /// If `FontId(0)` has not been defined yet, this sets the default font.
+    ///
+    /// This *must* be called (at least once) before any other font selection
     /// method, and before querying any font-derived properties (such as text
     /// dimensions).
     #[inline]
     pub fn select_default(&self) -> Result<(), Box<dyn std::error::Error>> {
+        self.db.write().unwrap().init();
         if self.fonts.read().unwrap().fonts.is_empty() {
             let id = self.select_font(&FontSelector::default())?;
             debug_assert!(id == FontId::default());
